from logging import getLogger
from typing import Type
from ..loaders import FileLoader, YamlLoader
from ..patterns import PatternSet, PatternSetData, PatternSetRepository
from ..Repository import Repository
from ..utils import Filepath, singleton
from .Detector import Detector


logger = getLogger(__name__)


@singleton
class DetectorRepository(Repository):
    def __init__(self, file_loader: Type[FileLoader] = YamlLoader):
        super().__init__(Detector)
        self.file_loader = file_loader()
        self.pattern_set_repo = PatternSetRepository()

    def cache_key(self, pattern_set: PatternSet) -> str:
        return pattern_set.name

    def create(self, pattern_set_filepath: str) -> Detector:
        """Create and cache a Detector instance."""
        fp = Filepath(pattern_set_filepath)
        pset_data: PatternSetData = self.file_loader(fp.filepath)
        pattern_set: PatternSet = self.pattern_set_repo.create(fp.name, pset_data)
        return super().create(pattern_set)



from logging import getLogger
from functools import wraps


logger = getLogger(__name__)


def singleton(klass):
    instances = []

    @wraps(klass.__new__)
    def _singleton(*args, **kwargs):
        if not instances:
            logger.debug(f"Constructing a new {klass}")
            instance = klass(*args, **kwargs)
            instances.append(instance)
        else:
            logger.debug(f"Retrieving an existing {klass}")
        return instances[0]

    return _singleton



from logging import getLogger
from ..patterns import Pattern, PatternSet
from .MatchSetMatcher import MatchSetMatcher


logger = getLogger(__name__)


class PatternSetMatcher(MatchSetMatcher):
    """A callable `Matcher` that adds the `Pattern`s of the given `PatternSet` to the underlying matcher."""

    def __init__(self, pattern_set: PatternSet) -> None:
        logger.debug("Constructing the PatternSetMatcher")
        super().__init__(
            best_match=pattern_set.best_match,
            how_many_matches=pattern_set.how_many_matches,
        )
        self.pset: PatternSet = pattern_set

        logger.info(
            f"Adding {len(self.pset.patterns)} Patterns to '{self.pset.name}' PatternSetMatcher"
        )
        for pattern_key in self.pset.patterns:
            pattern: Pattern = self.pset.patterns[pattern_key]
            spacy_config: dict[str, str] = {"greedy": "LONGEST"}
            logger.debug(f"Adding the '{pattern.rulename}' Pattern")
            self._matcher.add(pattern.rulename, [pattern.tokens], **spacy_config)



from logging import getLogger
from typing import Literal, Union
from ..Config import Config
from ..defaults import PATTERN_SET_DEFAULT_BEST_MATCH, PATTERN_SET_DEFAULT_HOW_MANY_MATCHES
from .Pattern import Pattern, PatternData


MetaSettingKeys = Literal[
    "best_match",
    "extract_noun_chunks",
    "how_many_matches",
    "skip_tests",
]
MetaSettingValues = Union[str, bool]
Meta = dict[MetaSettingKeys, MetaSettingValues]

TestKeys = Literal[
    "input",
    "rulenames",
    "skip",
    "spans",
]
TestInput = str
TestExpectedRulenames = list[str]
TestExpectedSpans = list[str]
Test = dict[TestKeys, Union[TestInput, TestExpectedRulenames, TestExpectedSpans]]

PatternSetData = dict[str, Union[list[PatternData], Meta, list[Test]]]


logger = getLogger(__name__)


class PatternSet:
    def __init__(self, name: str, data: PatternSetData) -> None:
        logger.debug(f"Constructing the '{name}' PatternSet")
        self.name: str = name

        self.patterns: dict[str, Pattern] = {}
        for p_data in data["patterns"]:
            pattern = Pattern(p_data)
            logger.info(f"Adding '{pattern.rulename}' to '{self.name}'")
            self.patterns[pattern.rulename] = pattern

        self.meta: Meta = data["config"] if "config" in data else {}
        for meta_key in self.meta:
            logger.debug(f"Found '{self.meta[meta_key]}' for meta '{meta_key}'")

        self.tests: list[Test] = data["tests"] if "tests" in data else []
        for test in self.tests:
            logger.debug(f"Found test: {test['input']}")

    @property
    def best_match(self) -> str:
        try:
            return str(self.meta["best_match"])
        except:
            return PATTERN_SET_DEFAULT_BEST_MATCH

    @property
    def how_many_matches(self) -> str:
        try:
            return str(self.meta["how_many_matches"])
        except:
            return PATTERN_SET_DEFAULT_HOW_MANY_MATCHES

    @property
    def should_extract_noun_chunks(self) -> bool:
        try:
            return bool(self.meta["extract_noun_chunks"])
        except:
            return False

    @property
    def skip_tests(self) -> str:
        try:
            return str(self.meta["skip_tests"])
        except:
            return ""



from logging import getLogger
from spacy import explain
from spacy.language import Language
from tabulate import tabulate
from typing import Literal


logger = getLogger(__name__)


TokenlikeKeys = Literal["pos", "tag", "dep", "lemma", "word"]
Tokenlike = dict[TokenlikeKeys, str]


def token_data(
    nlp: Language,
    sentence: str,
    pos: bool = True,
    tag: bool = True,
    dependency: bool = True,
    lemma: bool = False,
    word: bool = False,
) -> list[Tokenlike]:
    """Generate a list of objects containing the parts-of-speech, tags, dependencies, and lemmas of each token. This is particularly helpful for writing new patterns.

    Keyword arguments:
    nlp         -- (Language) The `nlp` instance
    input       -- (str) The sentence or chunk of text to be analyzed
    pos         -- (bool) If True, include the part-of-speech property (default True)
    tag         -- (bool) If True, include the tag property (default True)
    dependency  -- (bool) If True, include the dependency property (default True)
    lemma       -- (bool) If True, include the lemma property (default False)
    word        -- (bool) If True, include the word itself (default False)
    """
    logger.debug(f"Extracting the token data for `{sentence}`")

    logger.debug("Tokenizing")
    tagged_words = nlp(sentence)

    logger.debug("Extracting")
    data: list[Tokenlike] = []
    for tagged_word in tagged_words:
        entry: Tokenlike = {}
        if pos:
            entry["pos"] = tagged_word.pos_
        if tag:
            entry["tag"] = tagged_word.tag_
        if dependency:
            entry["dep"] = tagged_word.dep_
        if lemma:
            entry["lemma"] = tagged_word.lemma_
        if word:
            entry["word"] = tagged_word.text
        data.append(entry)
    logger.debug("Returning the token data")
    return data



from logging import getLogger
from spacy.tokens import Doc
from ..Config import Config
from ..matches import MatchSet, RawMatch
from .RawMatcher import RawMatcher
from ..patterns import PATTERN_SET_DEFAULT_BEST_MATCH, PATTERN_SET_DEFAULT_HOW_MANY_MATCHES


logger = getLogger(__name__)


class MatchSetMatcher(RawMatcher):
    def __init__(
        self,
        best_match: str = PATTERN_SET_DEFAULT_BEST_MATCH,
        how_many_matches: str = PATTERN_SET_DEFAULT_HOW_MANY_MATCHES,
    ) -> None:
        logger.debug("Constructing the MatchSetMatcher")
        super().__init__()
        self.best_match = best_match
        self.how_many_matches = how_many_matches

    def __call__(self, doc: Doc) -> MatchSet:
        logger.debug(f"Calling the MatchSetMatcher on '{doc}'")
        raw_matches: list[RawMatch] = super().__call__(doc)
        return MatchSet(
            raw_matches,
            doc,
            best_match=self.best_match,
            how_many_matches=self.how_many_matches,
        )



from logging import basicConfig, Formatter, INFO, StreamHandler, WARNING
from sys import stderr, stdout
from .Config import Config
from .defaults import LOGGER_DEFAULT_LEVEL, LOGGER_FORMAT_BASIC, LOGGER_FORMAT_DETAILED


def configure_logger(log_level: int = LOGGER_DEFAULT_LEVEL) -> None:
    basic_formatter = Formatter(fmt=LOGGER_FORMAT_BASIC)
    detailed_formatter = Formatter(fmt=LOGGER_FORMAT_DETAILED)

    handler = StreamHandler(stdout)
    formatter: Formatter = basic_formatter if log_level == INFO else detailed_formatter
    handler.setFormatter(formatter)

    error_handler = StreamHandler(stderr)
    error_handler.setLevel(WARNING)
    error_handler.setFormatter(detailed_formatter)

    basicConfig(level=log_level, handlers=[handler, error_handler])



from logging import getLogger
from ..Repository import Repository
from ..utils import singleton
from .PatternSet import PatternSet, PatternSetData


logger = getLogger(__name__)


@singleton
class PatternSetRepository(Repository):
    def __init__(self):
        logger.debug("Constructing the PatternSetRepository")
        super().__init__(PatternSet)

    def cache_key(self, name: str, data: PatternSetData) -> str:
        return name

    def create(self, name: str, data: PatternSetData) -> PatternSet:
        """Create a new `PatternSet`, cache it, and return it. If a `PatternSet` with the same name has already been created, the cached instance will be returned."""
        logger.info(f"Creating or retrieving the '{name}' PatternSet")
        return super().create(name, data)



from .Pattern import Pattern
from .PatternSet import PatternSet, PatternSetData, PATTERN_SET_DEFAULT_BEST_MATCH, PATTERN_SET_DEFAULT_HOW_MANY_MATCHES
from .PatternSetRepository import PatternSetRepository



from logging import getLogger
from os import listdir
from os.path import abspath, join
from pathlib import Path
from typing import Any
from yaml import FullLoader, load as load_yaml
from .defaults import PATTERN_SET_FILE_EXTENSION, PATTERN_SET_HOST_DIR
from .utils import Filepath, singleton




logger = getLogger(__name__)


@singleton
class Config:
    """A class for managing configuration settings."""
    @property
    def project_root_path(self) -> str:
        """Returns the full path of the project's root directory."""
        p = abspath(__file__)
        return str(Path(p).parents[1])

    @property
    def internal_patternset_dirpath(self) -> str:
        """Returns the filepath to the directory containing the internal patternsets."""
        return join(self.project_root_path, PATTERN_SET_HOST_DIR)

    @property
    def internal_patternset_filenames(self) -> list[str]:
        """Returns a list of the internal patternsets' filenames with the extension."""
        filenames: list[str] = listdir(self.internal_patternset_dirpath)
        return [fn for fn in filenames if self.is_valid_patternset_filename(fn)]

    @property
    def internal_patternset_filepaths(self) -> list[str]:
        """Returns a list of the internal patternsets' filepaths."""
        filepaths = []
        for fname in self.internal_patternset_filenames:
            fpath = join(self.internal_patternset_dirpath, fname)
            filepaths.append(fpath)
        return filepaths

    def is_valid_patternset_filename(self, filename: str) -> bool:
        """Returns True if the filename is valid for patternset file. Otherwise, returns False."""
        logger.debug(f"Validating patternset filename '{filename}'")
        fp: Filepath = Filepath(filename)
        has_ext = bool(fp.extension == PATTERN_SET_FILE_EXTENSION)
        return bool(not fp.is_hidden and has_ext)



import unittest
from grammardetector.Nlp import Nlp


class TestGlobalNlpTokenizer(unittest.TestCase):
    def test_is_defined(self):
        nlp = Nlp()
        self.assertTrue(nlp)
        self.assertTrue(nlp._nlp)



from logging import getLogger
from ..Input import Input
from ..matchers import PatternSetMatcher
from ..matches import Match, MatchSet
from ..patterns import PatternSet
from ..utils import flatten


logger = getLogger(__name__)


class Detector:
    def __init__(self, pattern_set: PatternSet) -> None:
        logger.debug(f"Constructing the '{pattern_set.name}' Detector")
        self.pattern_set = pattern_set
        self.name = self.pattern_set.name

        logger.debug(f"Creating the '{self.pattern_set.name}' PatternSetMatcher")
        self.matcher = PatternSetMatcher(self.pattern_set)

    def __call__(self, raw: str) -> list[Match]:
        """The entrypoint for the Detector."""
        logger.info(f"Detecting '{self.name}'")
        # logger.info(f"Detecting for '{self.name}' feature in '{raw}'")
        input = Input(raw)
        input.extract_noun_chunks = bool(self.pattern_set.should_extract_noun_chunks)

        matches: list[list[Match]] = []
        for frag in input.fragments:
            match_set: MatchSet = self.matcher(frag)
            matches.append(match_set.result)
        return flatten(matches)



from logging import getLogger
from functools import cached_property
from os.path import split


logger = getLogger(__name__)


class Filepath:
    """A helper class for working with filepaths, especially with extracting a filepath's components."""

    def __init__(self, filepath: str) -> None:
        self._filepath: str = str(filepath)
        self.dir_path: str = ""
        self.filename: str = ""
        self.name: str = ""
        self.extension: str = ""

        (self.dir_path, self.filename) = split(self._filepath)
        fname_splits = self.filename.split(".")
        try:
            self.name = fname_splits[-2]  # Use -2 instead of 0 because hidden files's 0 is an empty string
        except:
            self.name = fname_splits[0]
        self.extension = fname_splits[-1]

    @property
    def filepath(self) -> str:
        """The full filepath, including the directories, name, and extension."""
        return self._filepath

    @filepath.setter
    def filepath(self, _: str) -> None:
        msg = f"The internal filepath attribute cannot be modified"
        logger.error(msg)
        raise AttributeError(msg)

    @cached_property
    def is_hidden(self) -> bool:
        """Returns True if the file is a hidden file. Otherwise, returns False."""
        return bool(self.filename.startswith("."))



from logging import getLogger
from spacy import explain
from spacy.language import Language
from tabulate import tabulate


logger = getLogger(__name__)


def token_table(
    nlp: Language,
    sentence: str,
    pos: bool = True,
    tag: bool = True,
    dependency: bool = True,
    lemma: bool = True,
) -> str:
    """Generate a table containing the parts-of-speech, tag, dependency, and lemma of each token. This is helpful for visualizing token sequences.

    Keyword arguments:
    input       -- (str) The sentence or chunk of text to be analyzed
    pos         -- (bool) If True, include the part-of-speech property (default True)
    tag         -- (bool) If True, include the tag property (default True)
    dependency  -- (bool) If True, include the dependency property (default True)
    lemma       -- (bool) If True, include the lemma property (default True)
    """
    logger.debug(f"Creating the token table for `{sentence}`")

    logger.debug("Creating the table headers")
    headers = []
    headers.append("Word")
    if pos:
        headers.append("POS")
        headers.append("POS Definition")
    if tag:
        headers.append("Tag")
        headers.append("Tag Definition")
    if dependency:
        headers.append("Dep.")
        headers.append("Dep. Definition")
    if lemma:
        headers.append("Lemma.")

    logger.debug("Tokenizing")
    tagged_words = nlp(sentence)

    logger.debug("Extracting the features")
    data = []
    for word in tagged_words:
        entry = []
        entry.append(word.text)
        if pos:
            entry.append(word.pos_)
            entry.append(explain(word.pos_))  # type: ignore # explain is untyped
        if tag:
            entry.append(word.tag_)
            entry.append(explain(word.tag_))  # type: ignore # explain is untyped
        if dependency:
            entry.append(word.dep_)
            entry.append(explain(word.dep_))  # type: ignore # explain is untyped
        if lemma:
            entry.append(word.lemma_)
        data.append(entry)
    logger.debug("Returning the token table")
    return tabulate(data, headers=headers, tablefmt="github")



from unittest import TestCase
from grammardetector.detectors import DetectorRepository
from grammardetector.utils import Filepath
from grammardetector.Config import Config


class TestDetectorTests(TestCase):
    """The test suite for the YAML patternset configuration files. When running this suite via `$ python -m unittest`, the tests in the YAML patternset configuration files will be run. This allows for the patterns to be tested by running the specified input then comparing the expected/actual matching rulenames and text spans. This information can then be used to make improvements to, expand upon, and battle harden the YAML patternset configuration files. As a general rule, when patternsets do not behave as expected, the solution is to improve the patterns or add new variations of existing patterns. The solution is rarely to add new features via code, so fixing errors in the patterns themselves should be the first step."""

    @classmethod
    def setUpClass(self):
        """Load the patternset configuration files to extract the tests."""
        self.config = Config()
        self.repo = DetectorRepository()
        for fpath in self.config.internal_patternset_filepaths:
            fp = Filepath(fpath)
            self.repo.create(fp.filepath)

    def should_skip_pset(self, pset):
        """Determine if the patternset's tests should be skipped. This operates at the patternset level."""
        should_skip = None
        skip_reason = None

        try:
            skip_setting = pset.skip_tests
            if type(skip_setting) == bool:
                should_skip = skip_setting
                skip_reason = ""
            elif type(skip_setting) == str:
                should_skip = bool(len(skip_setting))
                skip_reason = skip_setting
            else:
                msg = f"Expected a bool or str but got '{type(skip_setting)}'"
                raise TypeError(msg)
        except KeyError:
            should_skip = False
            skip_reason = ""
        except Exception as e:
            raise e

        return (should_skip, skip_reason)

    def should_skip_test(self, test):
        """Determine if the patternset's tests should be skipped. This operates at the test level."""
        should_skip = None
        skip_reason = None

        try:
            skip_setting = test["skip"]
            if type(skip_setting) == bool:
                should_skip = skip_setting
                skip_reason = ""
            elif type(skip_setting) == str:
                should_skip = bool(len(skip_setting))
                skip_reason = skip_setting
            else:
                msg = f"Expected a bool or str but got '{type(skip_setting)}'"
                raise TypeError(msg)
            return (should_skip, skip_reason)
        except KeyError:
            should_skip = False
            skip_reason = ""
        except Exception as e:
            raise e

        return (should_skip, skip_reason)

    def get_input(self, test):
        """Extract the input from the test entry."""
        return test["input"]

    def get_expected_results(self, test, pset):
        """Extract the expected rulenames and text spans from the test entry."""
        expected_rulenames = test["rulenames"] if "rulenames" in test else []
        expected_spans = test["spans"] if "spans" in test else []

        # Validate the presence of at least one key containing expected results
        if not expected_rulenames and not expected_spans:
            valid_keys = ["rulenames", "spans"]
            err_msg = f"The test must have at least one of these keys: {valid_keys}"
            raise KeyError(err_msg)

        # Validate the length of expected results in relation to the expected number of matches
        # expects_many_rulenames = bool(len(expected_rulenames) > 1)
        # expects_many_spans = bool(len(expected_spans) > 1)
        # if expects_many_rulenames or expects_many_spans:
        #     how_many_matches = pset.meta["how_many_matches"]
        #     one_match_setting = Config().prop_str("PATTERN_SET_ONE_MATCH")

        #     if how_many_matches.upper() == one_match_setting.upper():
        #         err_msg = "The patternset expects only one match, but the test expects multiple matches"
        #         raise ValueError(err_msg)
        return (expected_rulenames, expected_spans)

    def run_test(self, test, pset):
        """The main entrypoint to run an individual test."""
        input = self.get_input(test)

        # Determine if the test should be skipped
        (should_skip, skip_reason) = self.should_skip_test(test)
        if should_skip:
            msg = f"Skipping patternset '{pset.name}': '{input}'"
            if bool(len(skip_reason)):
                msg = msg + f" ({skip_reason})"
                print(msg)
            return

        # Extract expected results, run matcher, and compare to actual results
        # Separate each test using subTest to prevent failures from stopping other tests
        with self.subTest(f"{pset.name}:{input}"):
            (expected_rulenames, expected_spans) = self.get_expected_results(test, pset)
            rulenames = []
            spans = []

            detector = self.repo.get_one(pset.name)
            for match in detector(input):
                rulenames.append(match.rulename)
                spans.append(str(match.span))

            if expected_rulenames:
                self.assertListEqual(rulenames, expected_rulenames)
            if expected_spans:
                self.assertListEqual(spans, expected_spans)

    def run_detector(self, detector):
        """The main entrypoint to run all of the tests in a patternset."""
        pset = detector.pattern_set
        (should_skip, skip_reason) = self.should_skip_pset(pset)
        if should_skip:
            msg = f"Skipping patternset '{pset.name}'"
            if bool(len(skip_reason)):
                msg = msg + f": {skip_reason}"
                print(msg)
            return

        # Separate using subTest to prevent failures from stopping the run
        if pset.tests:
            with self.subTest(pset.name):
                for test in pset.tests:
                    self.run_test(test, pset)

    def test_pattern_set_json_tests(self):
        """The main entrypoint for the unittest package."""
        [self.run_detector(detector) for detector in self.repo.get_all()]
